
class chunk {
    static chunk const baseline_table[56];
	static chunk const star_table[56];
    static chunk const octonion_table[8][8];

    static constexpr bool subscript_range (int const n) {
        return (n >= 0) && (n <= 7);
    }

    static bool sort (int & a, int & b, int & c) {
        bool result {false};       
        if (a > b) { std::swap (a, b); result = ! result; }
        if (b > c) { std::swap (b, c); result = ! result; }
        if (a > b) { std::swap (a, b); result = ! result; }
        return result;
    }

    static int calc_abc (int const a, int const b, int const c) {
        int const part_a {(((a - 18) * a + 107) * a + 126) / 6};
        int const part_b {((- b + 13) * b - 56) / 2};
        int const part_c {c - 1};
        return part_a + part_b + part_c;
    }
public:
	int const sgn, vec;

	chunk (int const s, int const v) : sgn {s}, vec {v} {
        if (sgn > +1) throw fail {"chunk::chunk ... sgn > +1"};
        if (sgn < -1) throw fail {"chunk::chunk ... sgn < -1"};
        if (vec > +7) throw fail {"chunk::chunk ... vec > +7"};
        if (sgn == 0) {
            if (vec != -1) throw fail {"chunk::chunk ... sgn == 0, vec != -1"};
        } else {
            if (vec < 0) throw fail {"chunk::chunk ... sgn != 0, vec < 0"};
        }
    }

    static chunk const baseline (int a, int b, int c) {
        if (! subscript_range (a)) throw fail {"chunk::cross ... a out of range"};
        if (! subscript_range (b)) throw fail {"chunk::cross ... b out of range"};
        if (! subscript_range (c)) throw fail {"chunk::cross ... c out of range"};
        if (a == b) return ZZ;
        if (b == c) return ZZ;
        if (c == a) return ZZ;

        bool const negate {sort (a, b, c)};
        if (negate)
            return (- baseline_table [calc_abc (a, b, c)]);
        else
            return (+ baseline_table [calc_abc (a, b, c)]);
    }

    static chunk const star (int a, int b, int c) {
        if (! subscript_range (a)) throw fail {"chunk::star ... a out of range"};
        if (! subscript_range (b)) throw fail {"chunk::star ... b out of range"};
        if (! subscript_range (c)) throw fail {"chunk::star ... c out of range"};
        if (a == b) return ZZ;
        if (b == c) return ZZ;
        if (c == a) return ZZ;

        bool const negate {sort (a, b, c)};
        if (negate)
            return (- star_table [calc_abc (a, b, c)]);
        else
            return (+ star_table [calc_abc (a, b, c)]);
    }

    static chunk const octonion (int a, int b) {
        if (! subscript_range (a)) throw fail {"chunk::octonion ... a out of range"};
        if (! subscript_range (b)) throw fail {"chunk::octonion ... b out of range"};
        return octonion_table[a][b];
    }

	static chunk const ZZ;
	static chunk const P0, P1, P2, P3, P4, P5, P6, P7;
	static chunk const N0, N1, N2, N3, N4, N5, N6, N7;

    chunk operator+ () const { return chunk (+sgn, vec); }
    chunk operator- () const { return chunk (-sgn, vec); }

    bool operator== (chunk const & o) const {
        if (sgn != o.sgn) return false;
        if (vec != o.vec) return false;
        return true;
    }

    bool operator!= (chunk const & o) const {
        if (sgn != o.sgn) return true;
        if (vec != o.vec) return true;
        return false;
    }

    char const * const string () const {
        if (sgn == +1) {
            static char const * const p[8]
                {"P0", "P1", "P2", "P3", "P4", "P5", "P6", "P7", };
            return p[vec];
        }
        if (sgn == -1) {
            static char const * const n[8]
                {"N0", "N1", "N2", "N3", "N4", "N5", "N6", "N7", };
            return n[vec];
        }
        static char const * const z {"ZZ"};
        return z;
    }

    char const * const html_U () const {
        if (sgn == +1) {
            static char const * const p[8] {
                "+<i>U</i><sub>0</sub>",
                "+<i>U</i><sub>1</sub>",
                "+<i>U</i><sub>2</sub>",
                "+<i>U</i><sub>3</sub>",
                "+<i>U</i><sub>4</sub>",
                "+<i>U</i><sub>5</sub>",
                "+<i>U</i><sub>6</sub>",
                "+<i>U</i><sub>7</sub>"
            };
            return p[vec];
        }
        if (sgn == -1) {
            static char const * const n[8] {
                "&minus;<i>U</i><sub>0</sub>",
                "&minus;<i>U</i><sub>1</sub>",
                "&minus;<i>U</i><sub>2</sub>",
                "&minus;<i>U</i><sub>3</sub>",
                "&minus;<i>U</i><sub>4</sub>",
                "&minus;<i>U</i><sub>5</sub>",
                "&minus;<i>U</i><sub>6</sub>",
                "&minus;<i>U</i><sub>7</sub>",
            };
            return n[vec];
        }
        static char const * const z {"0"};
        return z;
    }
};

// these seventeen chunks are the only legal values:

chunk const chunk::ZZ {0, -1};

chunk const chunk::P0 {+1, 0};
chunk const chunk::P1 {+1, 1};
chunk const chunk::P2 {+1, 2};
chunk const chunk::P3 {+1, 3};
chunk const chunk::P4 {+1, 4};
chunk const chunk::P5 {+1, 5};
chunk const chunk::P6 {+1, 6};
chunk const chunk::P7 {+1, 7};

chunk const chunk::N0 {-1, 0};
chunk const chunk::N1 {-1, 1};
chunk const chunk::N2 {-1, 2};
chunk const chunk::N3 {-1, 3};
chunk const chunk::N4 {-1, 4};
chunk const chunk::N5 {-1, 5};
chunk const chunk::N6 {-1, 6};
chunk const chunk::N7 {-1, 7};

ostream & operator<< (ostream & o, chunk const & c)
{ o << c.string(); return o; }

chunk const chunk::baseline_table[56] = {
//  this is the function shown at
//  http://home.comcast.net/~tamivox/redbear/cross_3_8/full_table.html
//  and is a genuine cross product

//  the comments are abc combinations 

//  012 013 014 015 016 017 023 024 025 026 027
    P3, N2, P5, N4, P7, N6, P1, P6, N7, N4, P5,

//  034 035 036 037 045 046 047 056 057 067
    N7, N6, P5, P4, P1, P2, N3, N3, N2, P1,

//  123 124 125 126 127 134 135 136 137
    N0, P7, P6, N5, N4, P6, N7, N4, P5,

//  145 146 147 156 157 167
    N0, P3, P2, P2, N3, N0,

//  234 235 236 237 245 246 247 256 257 267
    N5, P4, N7, P6, N3, N0, N1, N1, P0, N3,

//  345 346 347 356 357 367 456 457 467 567
    P2, N1, P0, P0, P1, P2, N7, P6, N5, P4
};

chunk const chunk::star_table[56] = {
//  this an interesting function,
//  but it is not a cross product

//  the comments are abc combinations

//  012 013 014 015 016 017 023 024 025 026 027
    P4, N5, P6, N7, P2, N3, P6, N7, P3, P5, N1,

//  034 035 036 037 045 046 047 056 057 067
    P1, P4, N7, N2, P2, N3, P5, P1, P6, P4,

//  123 124 125 126 127 134 135 136 137
    N7, P5, P0, N3, N6, P2, N6, N0, P4,

//  145 146 147 156 157 167
    N3, P7, P0, P4, N2, N5,

//  234 235 236 237 245 246 247 256 257 267
    P0, N1, P4, N5, P6, P1, N3, N7, N4, P0,

//  345 346 347 356 357 367 456 457 467 567
    N7, N5, N6, N2, P0, N1, P0, N1, P2, N3
};

// for octonion multiplication;
// observe that the real component
// is at row 3 and column 3
chunk const chunk::octonion_table[8][8] = { 
    { N3, P2, N1, P0, N7, P6, N5, P4 },
    { N2, N3, P0, P1, N6, N7, P4, P5 },
    { P1, N0, N3, P2, N5, P4, P7, N6 },
    { P0, P1, P2, P3, P4, P5, P6, P7 },
    { P7, P6, P5, P4, N3, N2, N1, N0 },
    { N6, P7, N4, P5, P2, N3, P0, N1 },
    { P5, N4, N7, P6, P1, N0, N3, P2 },
    { N4, N5, P6, P7, P0, P1, N2, N3 }
};


