
namespace ident {

    void ortho (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        static eight_dim const ABC (oper (A, B, C));
        cout << "\n\n orthogonality ...";
        cout << "\n   should be zero: ";
        cout << "\n     A . (A x B x C) = " << dot (A, ABC);
        cout << "\n     B . (A x B x C) = " << dot (B, ABC);
        cout << "\n     C . (A x B x C) = " << dot (C, ABC);
        cout << "\n     (A x B x C) . A = " << dot (ABC, A);
        cout << "\n     (A x B x C) . B = " << dot (ABC, B);
        cout << "\n     (A x B x C) . C = " << dot (ABC, C);
    }

    void gram_det (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        static eight_dim const ABC (oper (A, B, C));
        cout << "\n\n gram determinant ...";
        cout << "\n   should be equal:";
        cout << "\n     (A x B x C) . (A x B x C) = " << dot (ABC, ABC);
        cout << "\n     gram determinant          = " << gram (A, B, C);
    }

    void anti_com (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n anti-commutativity ...";
        cout << "\n   should be equal:";
        cout << "\n     + (A x B x C) = " << + oper (A, B, C);
        cout << "\n     + (B x C x A) = " << + oper (B, C, A);
        cout << "\n     + (C x A x B) = " << + oper (C, A, B);
        cout << "\n     - (A x C x B) = " << - oper (A, C, B);
        cout << "\n     - (B x A x C) = " << - oper (B, A, C);
        cout << "\n     - (C x B x A) = " << - oper (C, B, A);
    }

    void trival_assoc (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n trivalued associativity ...";
        cout << "\n   should be equal:";
        cout << "\n     + ((A x  B x  C) x A  x B)  = " << + oper (oper (A, B, C), A, B);
        cout << "\n     -  (A x (B x  C  x A) x B)  = " << - oper (A, oper (B, C, A), B);
        cout << "\n     +  (A x  B x (C  x A  x B)) = " << + oper (A, B, oper (C, A, B));
    }

    void cycle (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n cycling ...";
        cout << "\n   should be equal:";
        cout << "\n     + ((A x B x C) . D) = " << + dot (oper (A, B, C), D);
        cout << "\n     - ((D x A x B) . C) = " << - dot (oper (D, A, B), C);
        cout << "\n     + ((C x D x A) . B) = " << + dot (oper (C, D, A), B);
        cout << "\n     - ((B x C x D) . A) = " << - dot (oper (B, C, D), A);
    }

    void distrib (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n distributivity ...";
        cout << "\n   should be equal:";
        cout << "\n     (A + B) x C x D           = " << oper ((A + B), C, D);
        cout << "\n     (A x C x D) + (B x C x D) = " << oper (A, C, D) + oper (B, C, D); 
        cout << "\n   should be equal:";
        cout << "\n     A x (B + C) x D           = " << oper (A, (B + C), D);
        cout << "\n     (A x B x D) + (A x C x D) = " << oper (A, B, D) + oper (A, C, D); 
        cout << "\n   should be equal:";
        cout << "\n     A x B x (C + D)           = " << oper (A, B, (C + D));
        cout << "\n     (A x B x C) + (A x B x D) = " << oper (A, B, C) + oper (A, B, D);
    }

    void scalar (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n factoring out scalars ...";
        cout << "\n   should be equal:";
        cout << "\n     (XA x YB x ZC) = " << oper (A * X, B * Y, C * Z);
        cout << "\n     XYZ(A x B x C) = " << oper (A, B, C) * (X * Y * Z);
    }

    void repeat (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n repeated factors ...";
        cout << "\n   should be zero:";
        cout << "\n     A x A x B = " << + oper (A, A, B);
        cout << "\n     A x B x A = " << + oper (A, B, A);
        cout << "\n     B x A x A = " << + oper (B, A, A);
    }

    void ortho_mag (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n ortho_mag ...";

        eight_dim const rand_P {eight_dim::random()};
        eight_dim const rand_Q {eight_dim::random()};
        eight_dim const rand_R {eight_dim::random()};

        eight_dim const P {rand_P
        };
        eight_dim const Q {rand_Q
            - P * ((dot (rand_Q, ~P) / P.squ_mag()))
        };
        eight_dim const R {rand_R
            - P * ((dot (rand_R, ~P) / P.squ_mag()))
            - Q * ((dot (rand_R, ~Q) / Q.squ_mag()))
        };
        eight_dim const S {oper (P, Q, R)};

        double const P_mag {P.mag()};
        double const Q_mag {Q.mag()};
        double const R_mag {R.mag()};
        double const S_mag {S.mag()};

        cout << "\n   define some arbitrary orthogonal vectors:";
        cout << "\n     P = " << P;
        cout << "\n     Q = " << Q;
        cout << "\n     R = " << R;
        cout << "\n   | P | = " << P_mag;
        cout << "\n   | Q | = " << Q_mag;
        cout << "\n   | R | = " << R_mag;
        cout << "\n   P x Q x R = " << S;
        cout << "\n   should be zero:";
        cout << "\n     dot (P, ~Q) = " << dot (P, ~Q);
        cout << "\n     dot (Q, ~R) = " << dot (Q, ~R);
        cout << "\n     dot (R, ~P) = " << dot (R, ~P);

        if (complex::imag_zero)
            cout << "\n   should be equal:";
        else if (eight_dim::four_only)
            cout << "\n   should be equal:";
        else
            cout << "\n   should NOT be equal:";

        cout << "\n     | P x Q x R |     = " << S_mag;
        cout << "\n     | P | | Q | | R | = " << (P_mag * Q_mag * R_mag);
    }

    void slicing (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n slicing ...";
        cout << "\n   should be equal:";
        eight_dim T;
        for (int n {0}; n < 14; ++n)
            T += oper (A.slice(n), B.slice(n), C.slice(n));  
        cout << "\n     A x B x C  = " << oper (A, B, C);
        cout << "\n     by slicing = " << T;
    }

    void basic_conj (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n conjugation ...";
        cout << "\n   should be equal:";
        cout << "\n       (~A x ~B x ~C) = " <<   oper (~A, ~B, ~C);
        cout << "\n     ~ ( A x  B x  C) = " << ~ oper ( A,  B,  C);
        cout << "\n   should be equal:";
        cout << "\n       (~A x  B x ~C) = " <<   oper (~A,  B, ~C);
        cout << "\n     ~ ( A x ~B x  C) = " << ~ oper ( A, ~B,  C);
    }

    void cycle_conj (
        eight_dim (* oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n conjugative cycling ...";
        cout << "\n   should be equal:";
        cout << "\n       + (( A x ~B x  C) . ~D) = " <<   + dot (oper (A, ~B, C), ~D);
        cout << "\n       + (( C x ~D x  A) . ~B) = " <<   + dot (oper (C, ~D, A), ~B);
        cout << "\n       - ((~B x  C x ~D) .  A) = " <<   - dot (oper (~B, C, ~D), A);
        cout << "\n       - ((~D x  A x ~B) .  C) = " <<   - dot (oper (~D, A, ~B), C);
        cout << "\n     ~ + ((~A x  B x ~C) .  D) = " << ~ + dot (oper (~A, B, ~C), D);
        cout << "\n     ~ + ((~C x  D x ~A) .  B) = " << ~ + dot (oper (~C, D, ~A), B);
        cout << "\n     ~ - (( B x ~C x  D) . ~A) = " << ~ - dot (oper (B, ~C, D), ~A);
        cout << "\n     ~ - (( D x ~A x  B) . ~C) = " << ~ - dot (oper (D, ~A, B), ~C);
    }

    void powers (
        eight_dim (* pow_oper) (
            eight_dim (*) (
                eight_dim const &, eight_dim const &, eight_dim const &
            ),
            eight_dim const &, eight_dim const &, eight_dim const &,
            int const
        ),

        eight_dim (* crs_oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        int const top {8};

        if (complex::imag_zero) {
            double prev_mag {0.0};
            
            for (int n {0}; n < top; n += 1) {
                cout << "\n   p = " << noshowpos << n << ": ";
                eight_dim const curr {pow_oper (crs_oper, A, B, C, n)};
                double const curr_mag {curr.mag()};
                cout << curr << noshowpos << "  mag: " << curr_mag;
                if (n > 1) {
                    cout << "   ratio: " << curr_mag << " / " << prev_mag;
                    cout << " = " << (curr_mag / prev_mag);
                }
                prev_mag = curr_mag;
            }
        } else {
            double prev_mag_0 {0.0};
            double prev_mag_1 {0.0};

            for (int n {0}; n < top; n += 2) {
                {
                    cout << "\n   p = " << noshowpos << n << ": ";
                    eight_dim const curr_0 {pow_oper (crs_oper, A, B, C, n)};
                    double const curr_mag_0 {curr_0.mag()};
                    cout << curr_0 << noshowpos << "  mag: " << curr_mag_0;
                    if (n > 2) {
                        cout << "   ratio: " << curr_mag_0 << " / " << prev_mag_0;
                        cout << " = " << (curr_mag_0 / prev_mag_0);
                    }
                    prev_mag_0 = curr_mag_0;
                } {
                    cout << "\n   p = " << noshowpos << (n + 1) << ": ";
                    eight_dim const curr_1 {pow_oper (crs_oper, A, B, C, n + 1)};
                    double const curr_mag_1 {curr_1.mag()};
                    cout << curr_1 << noshowpos << "  mag: " << curr_mag_1;
                    if (n > 1) {
                        cout << "   ratio: " << curr_mag_1 << " / " << prev_mag_1;
                        cout << " = " << (curr_mag_1 / prev_mag_1);
                    }
                    prev_mag_1 = curr_mag_1;
                }
            }
        }
    }

    void powers (
        eight_dim (* crs_oper) (
            eight_dim const &, eight_dim const &, eight_dim const &
        )
    ) {
        cout << "\n\n for instance A x B x (A x B x (A x B x C)) ...";
        if (complex::imag_zero) {
            cout << "\n   | A | | B | = " << A.mag() * B.mag();
            cout << "\n   A . B  = " << dot (A, B);
        }
        powers (power_sinisterior, crs_oper);

        cout << "\n\n for instance A x (A x (A x B x C) x C) x C ...";
        if (complex::imag_zero) {
            cout << "\n   | A | | C | = " << A.mag() * C.mag();
        }
        powers (power_exterior, crs_oper);

        cout << "\n\n for instance ((A x B x C) x B x C) x B x C ...";
        if (complex::imag_zero) {
            cout << "\n   | B | | C | = " << B.mag() * C.mag();
        }
        powers (power_dexterior, crs_oper);
    }
};
