
#include <iostream>
#include <vector>
#include <iomanip>
#include <cmath>
#include <random>
#include <string>
#include <cmath>

using std::cout;
using std::showpos;
using std::noshowpos;
using std::ostream;
using std::setw;
using std::vector;
using std::string;

struct fail {
   fail (char const * const m) 
      { cout << "\n\n *** fail *** " << m << " *** \n\n"; }   
};

struct rand_double {
    rand_double () :
        r (bind(
            std::uniform_real_distribution<>(-0.7,+0.7),
            std::default_random_engine()
        )) { }
    double operator()() const { return r(); }
private:
    std::function<double()> const r;
};

rand_double rd;

#include "chunk.h"
#include "complex.h"
#include "eight_dim_1.h"
#include "eight_dim_2.h"

// handy constants for demonstrations
eight_dim const A {eight_dim::random()};
eight_dim const B {eight_dim::random()};
eight_dim const C {eight_dim::random()};
eight_dim const D {eight_dim::random()};
eight_dim const E {eight_dim::random()};
complex   const X {complex::random()};
complex   const Y {complex::random()};
complex   const Z {complex::random()};

#include "ident.h"
#include "table.h"

int main () {
	cout.setf (std::ios::fixed, std::ios::floatfield);
	cout.precision (5);
    cout << showpos;

    eight_dim (* oper) (
        eight_dim const &, eight_dim const &, eight_dim const &
    )  =
    // choose one of these four cross products or supply your own:
    // &baseline;
    // &lee_leung;
    &gray_1;
    // &gray_2;

    // print a pretty table:
    // html_table (&baseline);
    // html_table (&lee_leung);
    // html_table (&gray_1);
    // html_table (&gray_2);

    if (true) {
        cout << "\n notation ...";
        cout << "\n   dot product of two vectors: A . B";
        cout << "\n   cross product of three vectors: A x B x C";
        cout << "\n   multiplication of vector A by scalar Z: ZA";
        if (! complex::imag_zero) cout << "\n   complex conjugate: ~A";

        cout << "\n\n define some arbitrary vectors ... ";
        cout << "\n   A = " << A;
        cout << "\n   B = " << B;
        cout << "\n   C = " << C;
        cout << "\n   D = " << D;
        cout << "\n   E = " << E;
        cout << "\n";
        cout << "\n   | A | = " << A.mag();
        cout << "\n   | B | = " << B.mag();
        cout << "\n   | C | = " << C.mag();
        cout << "\n   | D | = " << D.mag();
        cout << "\n   | E | = " << E.mag();

        cout << "\n\n define some arbitrary scalars ... ";
        cout << "\n   X = " << X;
        cout << "\n   Y = " << Y;
        cout << "\n   Z = " << Z;
    }

    if (true) {
        cout << "\n\n ===== essential cross product identities ===== ";
        ident::ortho (oper);
        ident::gram_det (oper);
    }

    if (true) {
        cout << "\n\n ===== other cross product identities ===== ";
        ident::anti_com (oper);
        ident::trival_assoc (oper);
        ident::cycle (oper);
        ident::distrib (oper);
        ident::scalar (oper);
        ident::repeat (oper);

        ident::ortho_mag (oper);

        if (! eight_dim::four_only)
            ident::slicing (oper);

        if (! complex::imag_zero) {
            cout << "\n\n ===== complex cross product identities ===== ";
            ident::basic_conj (oper);
            ident::cycle_conj (oper);
        }

        cout << "\n\n ===== powers ===== ";
        ident::powers (oper);
    }

    return 0;
}
